"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

var _eventstoreBase = require("@resolve-js/eventstore-base");

var _getLog = require("./get-log");

async function sleep(delay) {
  return new Promise(function (resolve) {
    setTimeout(resolve, delay);
  });
}

function statusDataToError(statusData, defaults) {
  const errorMessage = statusData != null && statusData.message != null ? statusData.message : defaults.message;
  const errorName = statusData != null && statusData.name != null ? statusData.name : defaults.name;
  const errorStack = statusData != null && statusData.stack != null ? statusData.stack : undefined;

  if (errorStack) {
    return {
      message: errorMessage,
      name: errorName,
      stack: errorStack
    };
  } else {
    return {
      message: errorMessage,
      name: errorName
    };
  }
}

const isHTTPServiceError = name => name === 'FetchError' || name === 'AbortError' || name === 'ServiceError';

const getBuildDelay = iterationNumber => {
  return 30000 * 2 ** iterationNumber;
};

const BATCH_PROCESSING_POLL_MS = 50;

const build = async (basePool, readModelName, store, modelInterop, next, eventstoreAdapter, getVacantTimeInMillis, buildInfo) => {
  var _buildInfo$iterationN;

  const log = (0, _getLog.getLog)('build');
  let iterationNumber = (_buildInfo$iterationN = buildInfo.iterationNumber) !== null && _buildInfo$iterationN !== void 0 ? _buildInfo$iterationN : 0;

  const delayNext = async (delay, error) => {
    log.debug(`Delaying next for ${delay}ms due to service error ${error ? error.name + ': ' + error.message : ''}`);
    await next(delay, {
      iterationNumber: iterationNumber + 1
    });
  };

  const state = await basePool.getReplicationState(basePool);

  if (state.status === 'error') {
    log.error(`Refuse to start or continue replication with error state: ${state.statusData}`);
    return;
  } else if (state.status === 'serviceError') {
    await delayNext(getBuildDelay(iterationNumber), state.statusData);
    return;
  }

  if (state.paused) {
    log.warn('Refuse to start or continue replication because it is paused');
    return;
  }

  const timeLeft = getVacantTimeInMillis();

  try {
    const result = await basePool.occupyReplication(basePool, timeLeft);

    if (!result.success) {
      log.error(`Could not occupy replication process: ${result.message}`);
      return;
    }
  } catch (error) {
    await delayNext(getBuildDelay(iterationNumber), error);
    return;
  }

  let iterator = state.iterator;
  let localContinue = true;
  await eventstoreAdapter.establishTimeLimit(getVacantTimeInMillis);
  let eventLoader;

  const onExit = async () => {
    try {
      if (eventLoader !== undefined) await eventLoader.close();
    } catch (error) {
      log.error(error);
    }

    try {
      await basePool.releaseReplication(basePool);
    } catch (error) {
      if (!isHTTPServiceError(error.name)) log.error(error);
    }
  };

  try {
    eventLoader = await eventstoreAdapter.getEventLoader({
      cursor: iterator == null ? null : iterator.cursor
    }, {
      preferRegular: basePool.preferRegularLoader
    });
    log.debug(eventLoader.isNative ? 'Using native event loader' : 'Using regular event loader');
  } catch (error) {
    await onExit();

    if (_eventstoreBase.RequestTimeoutError.is(error)) {
      await next();
    } else if (_eventstoreBase.ServiceBusyError.is(error)) {
      await delayNext(getBuildDelay(iterationNumber), error);
    } else {
      log.error(error);
    }

    return;
  }

  log.debug('Starting or continuing replication process');

  while (true) {
    let lastError = null;
    const cursor = iterator == null ? null : iterator.cursor;
    let loadEventsResult;
    let gatheredSecrets;

    try {
      loadEventsResult = await eventLoader.loadEvents(100);
      gatheredSecrets = await eventstoreAdapter.gatherSecretsFromEvents(loadEventsResult.events);
    } catch (error) {
      await onExit();

      if (_eventstoreBase.RequestTimeoutError.is(error)) {
        await next();
      } else if (_eventstoreBase.ServiceBusyError.is(error)) {
        await delayNext(getBuildDelay(iterationNumber), error);
      } else {
        log.error(error);
      }

      return;
    }

    const {
      cursor: nextCursor,
      events
    } = loadEventsResult;
    const {
      existingSecrets,
      deletedSecrets
    } = gatheredSecrets;
    let appliedEventsCount = 0;
    let wasPaused = false;

    try {
      log.verbose(`Calling replicate on ${events.length} events`);
      const result = await basePool.callReplicate(basePool, events, existingSecrets, deletedSecrets, {
        cursor
      });

      if (result.type === 'serverError') {
        lastError = {
          name: 'ServiceError',
          message: result.message
        };
      } else if (result.type === 'clientError') {
        lastError = {
          name: 'Error',
          message: result.message
        };
      } else if (result.type === 'launched') {
        while (true) {
          const state = await basePool.getReplicationState(basePool);

          if (state.status === 'batchInProgress') {
            await sleep(BATCH_PROCESSING_POLL_MS);
          } else if (state.status === 'batchDone') {
            iterator = {
              cursor: nextCursor
            };
            appliedEventsCount = state.statusData != null ? state.statusData.appliedEventsCount : 0;
            wasPaused = state.paused;
            iterationNumber = 0;
            break;
          } else if (state.status === 'serviceError') {
            lastError = statusDataToError(state.statusData, {
              message: 'Unknown service error',
              name: 'ServiceError'
            });
            break;
          } else if (state.status === 'error') {
            lastError = statusDataToError(state.statusData, {
              message: 'Unknown error',
              name: 'Error'
            });
            break;
          }
        }
      } else {
        const errorMessage = `Unhandled replicate result. HTTP status code: ${result.httpStatus}. Message: ${result.message}`;
        lastError = {
          name: 'Error',
          message: errorMessage
        };
      }
    } catch (error) {
      lastError = error;
    }

    if (appliedEventsCount > 0) {
      log.verbose(`Replicated batch of ${appliedEventsCount} events`);
    }

    let delay = 0;
    let shouldContinue = appliedEventsCount > 0;

    if (lastError) {
      if (isHTTPServiceError(lastError.name)) {
        delay = getBuildDelay(iterationNumber);
        shouldContinue = true;
        localContinue = false;
      } else {
        shouldContinue = false;
      }
    }

    if (getVacantTimeInMillis() < 0) {
      localContinue = false;
    }

    if (wasPaused) {
      log.debug('Pausing replication as requested');
      await onExit();
      return;
    }

    if (shouldContinue && localContinue && delay === 0) {
      log.verbose('Continuing replication in the local build loop');
    } else {
      await onExit();
      if (lastError) log.error(`Exiting replication loop due to error ${lastError.message}`);

      if (shouldContinue) {
        if (delay > 0) {
          await delayNext(delay, lastError);
        } else {
          await next();
        }
      }

      return;
    }
  }
};

var _default = build;
exports.default = _default;
//# sourceMappingURL=build.js.map